// xGenRecordID Package  1998, Deks
// Version info in readme.txt and help
//
// Deks Site: http://deks.webjump.com/
// E-mail: Deks@online.nsk.su
// ============================================================================
// Main unit for package.

unit uxGenRecordID;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DB, DBTables;

const
  XGID_ZERO_VALUE = 0; // Value, returned by GenerateID function if DataSet is empty
  XGID_INCREMENT = 1; // Increment for new ID 
  
type
  TxFieldName = string;

  // ==================
  // TxGenRecordID
  // ==================
  TxGenRecordID = class(TComponent)
  private
  protected
    FDataSet: TDBDataSet;
    FTableName: string;
    FKeyField: TxFieldName;
    FOnNewRecord: TDataSetNotifyEvent;

    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; AOperation: TOperation); override;

    procedure SetDataSet(Value: TDBDataSet);

    procedure SetHandler;
    procedure RestoreHandler;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function GenerateID: Integer;

    procedure NewRecordHandler(DataSet: TDataSet);
  published
    property DataSet: TDBDataSet read FDataSet write SetDataSet;
    property TableName: string read FTableName write FTableName;
    property KeyField: TxFieldName read FKeyField write FKeyField;
  end;

implementation

// ============================================================================
// TxGenRecordID
// ============================================================================
constructor TxGenRecordID.Create(AOwner: TComponent);
begin
  // Constructor for component.

  // call inherited constructor:
  inherited Create(AOwner);

  // init internal fields (variables):
  FDataSet := nil;
  FOnNewRecord := nil;
end;

// ----------------------------------------------------------------------------
destructor TxGenRecordID.Destroy;
begin
  // Destructor for component.

  // remove reference to this component from parent:
  if (FDataSet <> nil) {and (FDataSet.OnNewRecord = NewRecordHandler)} then
    FDataSet.OnNewRecord := FOnNewRecord;

  // clear reference to parent component:
  FDataSet := nil;

  // call inherited destructor:
  inherited Destroy;
end;

// ----------------------------------------------------------------------------
procedure TxGenRecordID.Loaded;
begin
  // Handle "after loading" event for component.

  // call inherited method:
  inherited Loaded;

  // set event handler after loading component:
  SetHandler;  
end;

// ----------------------------------------------------------------------------
procedure TxGenRecordID.Notification(AComponent: TComponent; AOperation: TOperation);
begin
  // Handle changing to components in container (maybe, form).

  // call inherited method:
  inherited Notification(AComponent, AOperation);

  // test links to other components and clear them, if they are broken by
  // current operation:
  if AOperation = opRemove then begin
    if AComponent = DataSet then DataSet := nil;
  end;
end;

// ----------------------------------------------------------------------------
procedure TxGenRecordID.SetDataSet(Value: TDBDataSet);
begin
  // Set DataSet property.

  // optimization - exit from procedure if old and new value for property
  // are identical:
  if Value = FDataSet then Exit;

  // restore old event handler for old DataSet (v1.2 bug fix):
  RestoreHandler; // v1.3 - calling procedure for restoring event handler
  
  // change value for data set:
  FDataSet := Value;

  // catch event handler for new DataSet (v1.2 bug fix):
  SetHandler;
end;

// ----------------------------------------------------------------------------
procedure TxGenRecordID.SetHandler;
var
  aDataSetHandler: TDataSetNotifyEvent;
  aComponentHandler: TDataSetNotifyEvent;
begin
  // Procedure to correctly set event handler for DataSet.

  // test component/objects:
  if FDataSet = nil then Exit;
  if (csDesigning in ComponentState) then Exit; // no actions in design-time

  // copy current event handlers to variables:
  aDataSetHandler := FDataSet.OnNewRecord;
  aComponentHandler := NewREcordHandler;

  // second, perform some tests:
  if (@aDataSetHandler = @aComponentHandler) then
    FOnNewRecord := nil // no event handler from DataSet!
  else
    FOnNewRecord := aDataSetHandler; // remember DataSet's event handler.

  // set new event handler:
  FDataSet.OnNewRecord := aComponentHandler;
end;

// ----------------------------------------------------------------------------
procedure TxGenRecordID.RestoreHandler;
var
  aDataSetHandler: TDataSetNotifyEvent;
  aComponentHandler: TDataSetNotifyEvent;
begin
  // Procedure to correctly restore event handler for DataSet.

  // test component/objects:
  if FDataSet = nil then Exit;
  if (csDesigning in ComponentState) then Exit; // no actions in design-time
  
  // copy current event handlers to variables:
  aDataSetHandler := FDataSet.OnNewRecord;
  aComponentHandler := NewREcordHandler;

  // second, perform some tests:
  if (@aDataSetHandler = @aComponentHandler) then
    FDataSet.OnNewRecord := FOnNewRecord;

  // clear event handler for component:
  FOnNewRecord := nil;
end;

// ----------------------------------------------------------------------------
function TxGenRecordID.GenerateID: Integer;
var
  NewID: TQuery;
begin
  // Generate new primary key for DataSet.

  // if exception is occured, return some value allways:
  Result := XGID_ZERO_VALUE;
  
  // create query to retrieve max ID from DataSet:
  NewID := TQuery.Create(Self);
  try
    // init query's properties and open it:
    NewID.DatabaseName := DataSet.DatabaseName;
    NewID.SQL.Clear;

    // v1.3 bug fix (added table name to field name): 
    NewID.SQL.Add( 'SELECT max("' + TableName + '"'+'."' + KeyField + '") FROM "' + TableName + '"');
    NewID.Open;

    // if DataSet is empty, return zero value:
    if NewID.Fields[0].IsNull then
      Result := XGID_ZERO_VALUE
    else
      Result := NewID.Fields[0].AsInteger + XGID_INCREMENT; // return max value incremented by specified number
  finally
    // free query:
    NewID.Free;
  end;
end;

// ----------------------------------------------------------------------------
procedure TxGenRecordID.NewRecordHandler(DataSet: TDataSet);
var
  aComponentHandler: TDataSetNotifyEvent;
begin
  // Event handler for TDBDataSet.OnNewRecord event.

  // test all used fields (variables) and Exit if something is not OK:
  if not (DataSet is TDBDataSet) then Exit;
  if (DataSet as TDBDataSet) <> FDataSet then Exit;
  if KeyField = '' then Exit;
  if TableName = '' then Exit;

  // generate new primary key value (ID) for DataSet (v1.2 bug fix):
  DataSet.FieldByName(KeyField).AsInteger := GenerateID;

  // call trapped event handler (v1.2 bug fix):
  aComponentHandler := NewRecordHandler; // v1.3 bug fix: test for infinite loop when calling trapped event
  
  if (@FOnNewRecord <> nil) and (@aComponentHandler <> @FOnNewRecord) then
    FOnNewRecord(DataSet);
end;

end.
